@@ -187,7 +187,7 @@ static int set_net_apply_config(struct ctrl_cmd *cmd, void *data)
 	struct gsm_bts *bts;
 
 	llist_for_each_entry(bts, &net->bts_list, list) {
-		if (!is_ipaccess_bts(bts))
+		if (!is_ipa_abisip_bts(bts))
 			continue;
 
 		/*
@@ -332,7 +332,7 @@ static int set_net_rf_lock(struct ctrl_cmd *cmd, void *data)
 
 	cmd->reply = talloc_asprintf(cmd, "%u", locked);
 	if (!cmd->reply) {
-		cmd->reply = "OOM.";
+		cmd->reply = "OOM";
 		return CTRL_CMD_ERROR;
 	}
 
@@ -367,7 +367,7 @@ static int get_net_rf_states(struct ctrl_cmd *cmd, void *data)
 {
 	cmd->reply = bsc_rf_states_c(cmd);
 	if (!cmd->reply) {
-		cmd->reply = "OOM.";
+		cmd->reply = "OOM";
 		return CTRL_CMD_ERROR;
 	}
 	return CTRL_CMD_REPLY;
@@ -686,7 +686,7 @@ static int sccplite_asp_ctrl_cmd_send(struct osmo_ss7_asp *asp, struct ctrl_cmd
 	struct msgb *msg;
 
 	/* don't attempt to send CTRL on a non-SCCPlite ASP */
-	if (asp->cfg.proto != OSMO_SS7_ASP_PROT_IPA)
+	if (osmo_ss7_asp_get_proto(asp) != OSMO_SS7_ASP_PROT_IPA)
 		return 0;
 
 	msg = ctrl_cmd_make(cmd);
