@@ -50,6 +50,7 @@
 #include <osmocom/core/byteswap.h>
 #include <osmocom/bsc/lb.h>
 #include <osmocom/bsc/lcs_loc_req.h>
+#include <osmocom/bsc/vgcs_fsm.h>
 
 #define S(x)	(1 << (x))
 
@@ -80,7 +81,7 @@ static const struct value_string gscon_fsm_event_names[] = {
 	{GSCON_EV_MO_COMPL_L3, "MO_COMPL_L3"},
 	{GSCON_EV_A_CONN_CFM, "MO-CONNECT.cfm"},
 	{GSCON_EV_A_CLEAR_CMD, "CLEAR_CMD"},
-	{GSCON_EV_A_DISC_IND, "DISCONNET.ind"},
+	{GSCON_EV_A_DISC_IND, "DISCONNECT.ind"},
 	{GSCON_EV_A_COMMON_ID_IND, "COMMON_ID.ind"},
 	{GSCON_EV_ASSIGNMENT_START, "ASSIGNMENT_START"},
 	{GSCON_EV_ASSIGNMENT_END, "ASSIGNMENT_END"},
@@ -212,6 +213,12 @@ void gscon_fsm_wait_sccp_rlsd_onenter(struct osmo_fsm_inst *fi, uint32_t prev_st
 	if (conn->lcs.loc_req)
 		osmo_fsm_inst_dispatch(conn->lcs.loc_req->fi, LCS_LOC_REQ_EV_CONN_CLEAR, NULL);
 
+	if (conn->vgcs_call.fi)
+		osmo_fsm_inst_dispatch(conn->vgcs_call.fi, VGCS_EV_CLEANUP, NULL);
+
+	if (conn->vgcs_chan.fi)
+		osmo_fsm_inst_dispatch(conn->vgcs_chan.fi, VGCS_EV_CLEANUP, NULL);
+
 	gscon_release_lchans(conn, true, bsc_gsm48_rr_cause_from_gsm0808_cause(conn->clear_cause));
 	osmo_mgcpc_ep_clear(conn->user_plane.mgw_endpoint);
 
@@ -246,6 +253,8 @@ static void gscon_release_lchan(struct gsm_subscriber_connection *conn, struct g
 		conn->lchan = NULL;
 	if (conn->ho.fi && conn->ho.new_lchan == lchan)
 		conn->ho.new_lchan = NULL;
+	if (conn->vgcs_chan.new_lchan == lchan)
+		conn->vgcs_chan.new_lchan = NULL;
 	if (conn->assignment.new_lchan == lchan)
 		conn->assignment.new_lchan = NULL;
 	lchan_release(lchan, do_rr_release, err, cause_rr,
@@ -303,6 +312,8 @@ static int validate_initial_user_data(struct osmo_fsm_inst *fi, struct msgb *msg
 	switch (bssmap_type) {
 	case BSS_MAP_MSG_HANDOVER_RQST:
 	case BSS_MAP_MSG_PERFORM_LOCATION_RQST:
+	case BSS_MAP_MSG_VGCS_VBS_SETUP:
+	case BSS_MAP_MSG_VGCS_VBS_ASSIGNMENT_RQST:
 		return 0;
 
 	default:
@@ -343,6 +354,20 @@ static void handle_initial_user_data(struct osmo_fsm_inst *fi, struct msgb *msg)
 		lcs_loc_req_start(conn, msg);
 		return;
 
+	case BSS_MAP_MSG_VGCS_VBS_SETUP:
+		rate_ctr_inc(&conn->sccp.msc->msc_ctrs->ctr[MSC_CTR_BSSMAP_RX_DT1_VGCS_VBS_SETUP]);
+		/* VGCS: MSC asks vor voice group/bcast call. */
+		conn_fsm_state_chg(ST_ACTIVE);
+		vgcs_vbs_call_start(conn, msg);
+		return;
+
+	case BSS_MAP_MSG_VGCS_VBS_ASSIGNMENT_RQST:
+		rate_ctr_inc(&conn->sccp.msc->msc_ctrs->ctr[MSC_CTR_BSSMAP_RX_DT1_VGCS_VBS_ASSIGN_RQST]);
+		/* VGCS: MSC asks vor resource (channel) for voice group/bcast call. */
+		conn_fsm_state_chg(ST_ACTIVE);
+		vgcs_vbs_chan_start(conn, msg);
+		return;
+
 	default:
 		LOGPFSML(fi, LOGL_ERROR, "No support for initial BSSMAP: %s: %s\n",
 			 gsm0808_bssap_name(bs->type), gsm0808_bssmap_name(bssmap_type));
@@ -717,13 +742,19 @@ struct osmo_mgcpc_ep *gscon_ensure_mgw_endpoint(struct gsm_subscriber_connection
 			 msc_assigned_cic, osmo_mgcpc_ep_name(conn->user_plane.mgw_endpoint));
 
 	} else if (gscon_is_aoip(conn)) {
-		if (is_ipaccess_bts(for_lchan->ts->trx->bts))
+		if (is_ipa_abisip_bts(for_lchan->ts->trx->bts))
 			/* use dynamic RTPBRIDGE endpoint allocation in MGW */
 			epname = mgcp_client_rtpbridge_wildcard(mgcp_client);
 		else {
+			uint8_t i460_bit_offs;
+			if (for_lchan->ts->e1_link.e1_ts_ss == E1_SUBSLOT_FULL)
+				i460_bit_offs = 0;
+			else
+				i460_bit_offs = for_lchan->ts->e1_link.e1_ts_ss * 2;
+
 			epname = mgcp_client_e1_epname(conn, mgcp_client, for_lchan->ts->e1_link.e1_nr,
 						       for_lchan->ts->e1_link.e1_ts, 16,
-						       for_lchan->ts->e1_link.e1_ts_ss*2);
+						       i460_bit_offs);
 		}
 
 		conn->user_plane.mgw_endpoint =
@@ -763,7 +794,7 @@ bool gscon_connect_mgw_to_msc(struct gsm_subscriber_connection *conn,
 
 	mgw_info = (struct mgcp_conn_peer){
 		.port = port,
-		.call_id = conn->sccp.conn_id,
+		.call_id = conn->sccp.conn.conn_id,
 		.ptime = 20,
 		.x_osmo_osmux_use = conn->assignment.req.use_osmux,
 		.x_osmo_osmux_cid = conn->assignment.req.osmux_cid,
@@ -930,6 +961,10 @@ void gscon_lchan_releasing(struct gsm_subscriber_connection *conn, struct gsm_lc
 		if (conn->ho.fi)
 			osmo_fsm_inst_dispatch(conn->ho.fi, HO_EV_LCHAN_ERROR, lchan);
 	}
+	if (conn->vgcs_chan.new_lchan == lchan) {
+		if (conn->vgcs_chan.fi)
+			osmo_fsm_inst_dispatch(conn->vgcs_chan.fi, VGCS_EV_LCHAN_ERROR, lchan);
+	}
 	if (conn->lchan == lchan) {
 		lchan_forget_conn(conn->lchan);
 		conn->lchan = NULL;
@@ -966,6 +1001,10 @@ void gscon_forget_lchan(struct gsm_subscriber_connection *conn, struct gsm_lchan
 		conn->ho.new_lchan = NULL;
 		detach_label = "ho.new_lchan";
 	}
+	if (conn->vgcs_chan.new_lchan == lchan) {
+		conn->vgcs_chan.new_lchan = NULL;
+		detach_label = "vgcs.new_lchan";
+	}
 	if (conn->lchan == lchan) {
 		conn->lchan = NULL;
 		detach_label = "primary lchan";
@@ -984,6 +1023,7 @@ void gscon_forget_lchan(struct gsm_subscriber_connection *conn, struct gsm_lchan
 	if (!conn->lchan
 	    && !conn->ho.new_lchan
 	    && !conn->assignment.new_lchan
+	    && !conn->vgcs_chan.new_lchan
 	    && !conn->lcs.loc_req)
 		gscon_bssmap_clear(conn, GSM0808_CAUSE_EQUIPMENT_FAILURE);
 }
@@ -1005,6 +1045,7 @@ static void gscon_forget_mgw_endpoint(struct gsm_subscriber_connection *conn)
 	conn->ho.created_ci_for_msc = NULL;
 	lchan_forget_mgw_endpoint(conn->lchan);
 	lchan_forget_mgw_endpoint(conn->assignment.new_lchan);
+	lchan_forget_mgw_endpoint(conn->vgcs_chan.new_lchan);
 	lchan_forget_mgw_endpoint(conn->ho.new_lchan);
 }
 
@@ -1094,6 +1135,7 @@ static void gscon_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cau
 
 	lchan_forget_conn(conn->lchan);
 	lchan_forget_conn(conn->assignment.new_lchan);
+	lchan_forget_conn(conn->vgcs_chan.new_lchan);
 	lchan_forget_conn(conn->ho.new_lchan);
 
 	lb_close_conn(conn);
@@ -1102,9 +1144,14 @@ static void gscon_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cau
 		LOGPFSML(fi, LOGL_DEBUG, "Disconnecting SCCP\n");
 		struct bsc_msc_data *msc = conn->sccp.msc;
 		/* FIXME: include a proper cause value / error message? */
-		osmo_sccp_tx_disconn(msc->a.sccp_user, conn->sccp.conn_id, &msc->a.bsc_addr, 0);
+		osmo_sccp_tx_disconn(msc->a.sccp_user, conn->sccp.conn.conn_id, &msc->a.bsc_addr, 0);
 		conn->sccp.state = SUBSCR_SCCP_ST_NONE;
 	}
+	if (conn->sccp.conn.conn_id != SCCP_CONN_ID_UNSET && conn->sccp.msc) {
+		struct bsc_sccp_inst *bsc_sccp = osmo_sccp_get_priv(conn->sccp.msc->a.sccp);
+		bsc_sccp_inst_unregister_gscon(bsc_sccp, &conn->sccp.conn);
+		conn->sccp.conn.conn_id = SCCP_CONN_ID_UNSET;
+	}
 
 	if (conn->bsub) {
 		LOGPFSML(fi, LOGL_DEBUG, "Putting bsc_subscr\n");
@@ -1141,6 +1188,12 @@ static void gscon_pre_term(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause ca
 		conn->lcls.fi = NULL;
 	}
 
+	if (conn->vgcs_call.fi)
+		osmo_fsm_inst_dispatch(conn->vgcs_call.fi, VGCS_EV_CLEANUP, NULL);
+
+	if (conn->vgcs_chan.fi)
+		osmo_fsm_inst_dispatch(conn->vgcs_chan.fi, VGCS_EV_CLEANUP, NULL);
+
 	LOGPFSML(fi, LOGL_DEBUG, "Releasing all lchans (if any) because this conn is terminating\n");
 	gscon_release_lchans(conn, true, bsc_gsm48_rr_cause_from_gsm0808_cause(conn->clear_cause));
 
@@ -1202,7 +1255,7 @@ static struct osmo_fsm gscon_fsm = {
 	.event_names = gscon_fsm_event_names,
 };
 
-void bsc_subscr_conn_fsm_init(void)
+static __attribute__((constructor)) void bsc_subscr_conn_fsm_init(void)
 {
 	OSMO_ASSERT(osmo_fsm_register(&gscon_fsm) == 0);
 	OSMO_ASSERT(osmo_fsm_register(&lcls_fsm) == 0);
@@ -1220,8 +1273,8 @@ struct gsm_subscriber_connection *bsc_subscr_con_allocate(struct gsm_network *ne
 	conn->network = net;
 	INIT_LLIST_HEAD(&conn->dtap_queue);
 	INIT_LLIST_HEAD(&conn->hodec2.penalty_timers);
-	conn->sccp.conn_id = -1;
-
+	bscp_sccp_conn_node_init(&conn->sccp.conn, conn);
+	bscp_sccp_conn_node_init(&conn->lcs.lb.conn, conn);
 	/* Default clear cause (on RR translates to GSM48_RR_CAUSE_ABNORMAL_UNSPEC) */
 	conn->clear_cause = GSM0808_CAUSE_EQUIPMENT_FAILURE;
 
@@ -1394,7 +1447,7 @@ void gscon_update_id(struct gsm_subscriber_connection *conn)
 {
 	osmo_fsm_inst_update_id_f(conn->fi, "msc%u-conn%u%s%s",
 				  conn->sccp.msc ? conn->sccp.msc->nr : UINT_MAX,
-				  conn->sccp.conn_id,
+				  conn->sccp.conn.conn_id,
 				  conn->bsub? "_" : "",
 				  conn->bsub? bsc_subscr_id(conn->bsub) : "");
 }
