@@ -44,8 +44,9 @@
 #include <osmocom/bsc/gsm_04_08_rr.h>
 #include <osmocom/bsc/bts.h>
 #include <osmocom/bsc/bts_sm.h>
+#include <osmocom/bsc/timeslot_fsm.h>
 
-static int pcu_sock_send(struct gsm_bts *bts, struct msgb *msg);
+static int pcu_sock_send(struct gsm_network *net, struct msgb *msg);
 
 static const char *sapi_string[] = {
 	[PCU_IF_SAPI_RACH] =	"RACH",
@@ -54,14 +55,14 @@ static const char *sapi_string[] = {
 	[PCU_IF_SAPI_BCCH] =	"BCCH",
 	[PCU_IF_SAPI_PDTCH] =	"PDTCH",
 	[PCU_IF_SAPI_PRACH] =	"PRACH",
-	[PCU_IF_SAPI_PTCCH] = 	"PTCCH",
-	[PCU_IF_SAPI_AGCH_DT] = 	"AGCH_DT",
+	[PCU_IF_SAPI_PTCCH] =	"PTCCH",
+	[PCU_IF_SAPI_PCH_2] =	"PCH_2",
+	[PCU_IF_SAPI_AGCH_2] =	"AGCH_2",
 };
 
-/* Check if BTS has a PCU connection */
-static bool pcu_connected(struct gsm_bts *bts)
+bool pcu_connected(const struct gsm_network *net)
 {
-	struct pcu_sock_state *state = bts->pcu_state;
+	struct pcu_sock_state *state = net->pcu_state;
 
 	if (!state)
 		return false;
@@ -106,6 +107,19 @@ static bool ts_now_usable_as_pdch(const struct gsm_bts_trx_ts *ts)
 	}
 }
 
+/* Check if it is possible to use the TS as PDCH (not now, but maybe later) */
+static bool ts_usable_as_pdch(const struct gsm_bts_trx_ts *ts)
+{
+	switch (ts->pchan_from_config) {
+	case GSM_PCHAN_TCH_F_PDCH:
+	case GSM_PCHAN_OSMO_DYN:
+	case GSM_PCHAN_PDCH:
+		return true;
+	default:
+		return false;
+	}
+}
+
 /* Fill the frequency hopping parameter */
 static void info_ind_fill_fhp(struct gsm_pcu_if_info_trx_ts *ts_info,
 			      const struct gsm_bts_trx_ts *ts)
@@ -174,7 +188,7 @@ static int pcu_tx_info_ind(struct gsm_bts *bts)
 
 	bts_sm = bts->site_mgr;
 
-	LOGP(DPCU, LOGL_INFO, "Sending info for BTS %d\n", bts->nr);
+	LOG_BTS(bts, DPCU, LOGL_INFO, "Sending info for BTS\n");
 
 	rlcc = &bts->gprs.cell.rlc_cfg;
 
@@ -288,14 +302,105 @@ static int pcu_tx_info_ind(struct gsm_bts *bts)
 		info_ind_fill_trx(&info_ind->trx[trx->nr], trx);
 	}
 
-	return pcu_sock_send(bts, msg);
+	return pcu_sock_send(bts->network, msg);
+}
+
+static int pcu_tx_e1_ccu_ind(struct gsm_bts *bts)
+{
+	struct gsm_bts_trx *trx;
+
+	llist_for_each_entry(trx, &bts->trx_list, list) {
+		struct gsm_pcu_if_e1_ccu_ind *e1_ccu_ind;
+		int i;
+
+		if (trx->nr >= PCU_IF_NUM_TRX) {
+			LOG_TRX(trx, DPCU, LOGL_NOTICE, "PCU interface (version %u) "
+				"cannot handle more than %u transceivers => skipped\n",
+				PCU_IF_VERSION, PCU_IF_NUM_TRX);
+			continue;
+		}
+
+		for (i = 0; i < ARRAY_SIZE(trx->ts); i++) {
+			struct gsm_pcu_if *pcu_prim;
+			struct gsm_bts_trx_ts *ts;
+			struct msgb *msg;
+			int rc;
+
+			ts = &trx->ts[i];
+
+			if (ts->mo.nm_state.operational != NM_OPSTATE_ENABLED)
+				continue;
+			if (!ts_usable_as_pdch(ts))
+				continue;
+
+			msg = pcu_msgb_alloc(PCU_IF_MSG_E1_CCU_IND, bts->nr);
+			if (!msg)
+				return -ENOMEM;
+			pcu_prim = (struct gsm_pcu_if *)msg->data;
+			e1_ccu_ind = &pcu_prim->u.e1_ccu_ind;
+			e1_ccu_ind->ts_nr = ts->nr;
+			e1_ccu_ind->trx_nr = trx->nr;
+			e1_ccu_ind->e1_nr = ts->e1_link.e1_nr;
+			e1_ccu_ind->e1_ts = ts->e1_link.e1_ts;
+			e1_ccu_ind->e1_ts_ss = ts->e1_link.e1_ts_ss;
+
+			LOG_TRX(trx, DPCU, LOGL_INFO, "Sending E1 CCU info for TS %d\n", e1_ccu_ind->ts_nr);
+			rc = pcu_sock_send(bts->network, msg);
+			if (rc < 0)
+				return -EINVAL;
+		}
+	}
+
+	return 0;
 }
 
 /* Allow test to overwrite it */
 __attribute__((weak)) void pcu_info_update(struct gsm_bts *bts)
 {
-	if (pcu_connected(bts))
-		pcu_tx_info_ind(bts);
+	if (pcu_connected(bts->network)) {
+		if (bsc_co_located_pcu(bts)) {
+			/* In cases where the CCU is connected via an E1 line, we transmit the connection parameters for the
+			 * PDCH before we announce the other BTS related parameters. */
+			if (is_e1_bts(bts))
+				pcu_tx_e1_ccu_ind(bts);
+			pcu_tx_info_ind(bts);
+		}
+	}
+}
+
+static int pcu_tx_data_ind(struct gsm_bts_trx_ts *ts, uint8_t sapi, uint32_t fn,
+			   uint16_t arfcn, uint8_t block_nr, uint8_t *data, uint8_t len,
+			   int8_t rssi, uint16_t ber10k, int16_t bto, int16_t lqual)
+{
+	struct msgb *msg;
+	struct gsm_pcu_if *pcu_prim;
+	struct gsm_pcu_if_data *data_ind;
+	struct gsm_bts *bts = ts->trx->bts;
+
+	LOGP(DPCU, LOGL_DEBUG, "Sending data indication: sapi=%s arfcn=%d block=%d data=%s\n",
+	     sapi_string[sapi], arfcn, block_nr, osmo_hexdump(data, len));
+
+	msg = pcu_msgb_alloc(PCU_IF_MSG_DATA_IND, bts->nr);
+	if (!msg)
+		return -ENOMEM;
+	pcu_prim = (struct gsm_pcu_if *) msg->data;
+	data_ind = &pcu_prim->u.data_ind;
+
+	data_ind->sapi = sapi;
+	data_ind->rssi = rssi;
+	data_ind->fn = fn;
+	data_ind->arfcn = arfcn;
+	data_ind->trx_nr = ts->trx->nr;
+	data_ind->ts_nr = ts->nr;
+	data_ind->block_nr = block_nr;
+	data_ind->ber10k = ber10k;
+	data_ind->ta_offs_qbits = bto;
+	data_ind->lqual_cb = lqual;
+	if (len)
+		memcpy(data_ind->data, data, len);
+	data_ind->len = len;
+
+	return pcu_sock_send(bts->network, msg);
 }
 
 /* Forward rach indication to PCU */
@@ -307,13 +412,12 @@ int pcu_tx_rach_ind(struct gsm_bts *bts, int16_t qta, uint16_t ra, uint32_t fn,
 	struct gsm_pcu_if_rach_ind *rach_ind;
 
 	/* Bail if no PCU is connected */
-	if (!pcu_connected(bts)) {
-		LOGP(DRSL, LOGL_ERROR, "BTS %d CHAN RQD(GPRS) but PCU not "
-			"connected!\n", bts->nr);
+	if (!pcu_connected(bts->network)) {
+		LOG_BTS(bts, DRSL, LOGL_ERROR, "CHAN RQD(GPRS) but PCU not connected!\n");
 		return -ENODEV;
 	}
 
-	LOGP(DPCU, LOGL_INFO, "Sending RACH indication: qta=%d, ra=%d, "
+	LOG_BTS(bts, DPCU, LOGL_INFO, "Sending RACH indication: qta=%d, ra=%d, "
 		"fn=%d\n", qta, ra, fn);
 
 	msg = pcu_msgb_alloc(PCU_IF_MSG_RACH_IND, bts->nr);
@@ -329,37 +433,37 @@ int pcu_tx_rach_ind(struct gsm_bts *bts, int16_t qta, uint16_t ra, uint32_t fn,
 	rach_ind->is_11bit = is_11bit;
 	rach_ind->burst_type = burst_type;
 
-	return pcu_sock_send(bts, msg);
+	return pcu_sock_send(bts->network, msg);
 }
 
-/* Confirm the sending of an immediate assignment to the pcu */
-int pcu_tx_imm_ass_sent(struct gsm_bts *bts, uint32_t tlli)
+int pcu_tx_data_cnf(struct gsm_bts *bts, uint32_t msg_id, uint8_t sapi)
 {
 	struct msgb *msg;
 	struct gsm_pcu_if *pcu_prim;
-	struct gsm_pcu_if_data_cnf_dt *data_cnf_dt;
+	struct gsm_pcu_if_data_cnf *data_cnf;
 
-	LOGP(DPCU, LOGL_INFO, "Sending PCH confirm with direct TLLI\n");
+	LOGP(DPCU, LOGL_DEBUG, "Sending DATA.cnf: sapi=%s msg_id=%08x\n",
+	     sapi_string[sapi], msg_id);
 
-	msg = pcu_msgb_alloc(PCU_IF_MSG_DATA_CNF_DT, bts->nr);
+	msg = pcu_msgb_alloc(PCU_IF_MSG_DATA_CNF_2, bts->nr);
 	if (!msg)
 		return -ENOMEM;
 	pcu_prim = (struct gsm_pcu_if *) msg->data;
-	data_cnf_dt = &pcu_prim->u.data_cnf_dt;
+	data_cnf = &pcu_prim->u.data_cnf2;
 
-	data_cnf_dt->sapi = PCU_IF_SAPI_PCH;
-	data_cnf_dt->tlli = tlli;
+	data_cnf->sapi = sapi;
+	data_cnf->msg_id = msg_id;
 
-	return pcu_sock_send(bts, msg);
+	return pcu_sock_send(bts->network, msg);
 }
 
 /* we need to decode the raw RR paging message (see PCU code
  * Encoding::write_paging_request) and extract the mobile identity
  * (P-TMSI) from it */
-static int pcu_rx_rr_paging(struct gsm_bts *bts, uint8_t paging_group,
-			    const uint8_t *raw_rr_msg)
+static int pcu_rx_rr_paging_pch(struct gsm_bts *bts, uint8_t paging_group,
+				const struct gsm_pcu_if_pch *pch)
 {
-	struct gsm48_paging1 *p1 = (struct gsm48_paging1 *) raw_rr_msg;
+	struct gsm48_paging1 *p1 = (struct gsm48_paging1 *) pch->data;
 	uint8_t chan_needed;
 	struct osmo_mobile_identity mi;
 	int rc;
@@ -369,14 +473,14 @@ static int pcu_rx_rr_paging(struct gsm_bts *bts, uint8_t paging_group,
 		chan_needed = (p1->cneed2 << 2) | p1->cneed1;
 		rc = osmo_mobile_identity_decode(&mi, p1->data+1, p1->data[0], false);
 		if (rc) {
-			LOGP(DPCU, LOGL_ERROR, "PCU Sends paging "
-			     "request type %02x (chan_needed=%02x): Unable to decode Mobile Identity\n",
+			LOG_BTS(bts, DPCU, LOGL_ERROR, "PCU Sends paging "
+				"request type %02x (chan_needed=%02x): Unable to decode Mobile Identity\n",
 			     p1->msg_type, chan_needed);
 			rc = -EINVAL;
 			break;
 		}
-		LOGP(DPCU, LOGL_ERROR, "PCU Sends paging "
-		     "request type %02x (chan_needed=%02x, mi=%s)\n",
+		LOG_BTS(bts, DPCU, LOGL_ERROR, "PCU Sends paging "
+			"request type %02x (chan_needed=%02x, mi=%s)\n",
 		     p1->msg_type, chan_needed, osmo_mobile_identity_to_str_c(OTC_SELECT, &mi));
 		/* NOTE: We will have to add 2 to mi_len and subtract 2 from
 		 * the mi pointer because rsl_paging_cmd() will perform the
@@ -388,12 +492,12 @@ static int pcu_rx_rr_paging(struct gsm_bts *bts, uint8_t paging_group,
 		break;
 	case GSM48_MT_RR_PAG_REQ_2:
 	case GSM48_MT_RR_PAG_REQ_3:
-		LOGP(DPCU, LOGL_ERROR, "PCU Sends unsupported paging "
+		LOG_BTS(bts, DPCU, LOGL_ERROR, "PCU Sends unsupported paging "
 			"request type %02x\n", p1->msg_type);
 		rc = -EINVAL;
 		break;
 	default:
-		LOGP(DPCU, LOGL_ERROR, "PCU Sends unknown paging "
+		LOG_BTS(bts, DPCU, LOGL_ERROR, "PCU Sends unknown paging "
 			"request type %02x\n", p1->msg_type);
 		rc = -EINVAL;
 		break;
@@ -402,77 +506,76 @@ static int pcu_rx_rr_paging(struct gsm_bts *bts, uint8_t paging_group,
 	return rc;
 }
 
+static int pcu_rx_rr_imm_ass_pch(struct gsm_bts *bts, uint8_t paging_group,
+				 const struct gsm_pcu_if_pch *pch, bool confirm)
+{
+	LOG_BTS(bts, DPCU, LOGL_DEBUG, "PCU Sends immediate assignment via PCH (msg_id=0x%08x, IMSI=%s, Paging group=0x%02x)\n",
+		pch->msg_id, pch->imsi, paging_group);
+
+	/* NOTE: Sending an IMMEDIATE ASSIGNMENT via PCH became necessary with GPRS in order to be able to
+	 * assign downlink TBFs directly through the paging channel. However, this method never became part
+	 * of the RSL specs. This means that each BTS vendor has to come up with a proprietary method. At
+	 * the moment we only support Ericsson RBS here. */
+	if (is_ericsson_bts(bts))
+		return rsl_ericsson_imm_assign_cmd(bts, pch->msg_id, sizeof(pch->data), pch->data, paging_group,
+						   confirm);
+
+	LOG_BTS(bts, DPCU, LOGL_ERROR, "BTS model does not support sending immediate assignment via PCH!\n");
+	return -ENOTSUP;
+}
+
 static int pcu_rx_data_req(struct gsm_bts *bts, uint8_t msg_type,
 	struct gsm_pcu_if_data *data_req)
 {
-	struct msgb *msg;
-	char imsi_digit_buf[4];
-	uint32_t tlli = -1;
 	uint8_t pag_grp;
 	int rc = 0;
+	const struct gsm_pcu_if_pch *pch;
+	const struct gsm_pcu_if_agch *agch;
+	const struct gsm48_imm_ass *gsm48_imm_ass;
 
-	LOGP(DPCU, LOGL_DEBUG, "Data request received: sapi=%s arfcn=%d "
+	LOG_BTS(bts, DPCU, LOGL_DEBUG, "Data request received: sapi=%s arfcn=%d "
 		"block=%d data=%s\n", sapi_string[data_req->sapi],
 		data_req->arfcn, data_req->block_nr,
 		osmo_hexdump(data_req->data, data_req->len));
 
 	switch (data_req->sapi) {
-	case PCU_IF_SAPI_PCH:
-		/* the first three bytes are the last three digits of
-		 * the IMSI, which we need to compute the paging group */
-		imsi_digit_buf[0] = data_req->data[0];
-		imsi_digit_buf[1] = data_req->data[1];
-		imsi_digit_buf[2] = data_req->data[2];
-		imsi_digit_buf[3] = '\0';
-		LOGP(DPCU, LOGL_DEBUG, "SAPI PCH imsi %s\n", imsi_digit_buf);
-		pag_grp = gsm0502_calc_paging_group(&bts->si_common.chan_desc,
-						str_to_imsi(imsi_digit_buf));
-		pcu_rx_rr_paging(bts, pag_grp, data_req->data+3);
+	case PCU_IF_SAPI_AGCH_2:
+		if (data_req->len < sizeof(struct gsm_pcu_if_agch)) {
+			LOG_BTS(bts, DPCU, LOGL_ERROR, "Received PCU data request with invalid/small length %d\n",
+				data_req->len);
+			break;
+		}
+
+		agch = (struct gsm_pcu_if_agch *)data_req->data;
+		if (rsl_imm_assign_cmd(bts, GSM_MACBLOCK_LEN, agch->data))
+			return -EIO;
+
+		/* Send the confirmation immediately. This is as accurate as we can get since from this point on the
+		 * BTS hardware is responsible to schedule the sending of the IMMEDIATE ASSIGNMENT */
+		if (agch->confirm)
+			return pcu_tx_data_cnf(bts, agch->msg_id, PCU_IF_SAPI_AGCH_2);
 		break;
-	case PCU_IF_SAPI_AGCH:
-		msg = msgb_alloc(data_req->len, "pcu_agch");
-		if (!msg) {
-			rc = -ENOMEM;
+	case PCU_IF_SAPI_PCH_2:
+		if (data_req->len < sizeof(struct gsm_pcu_if_pch)) {
+			LOG_BTS(bts, DPCU, LOGL_ERROR, "Received PCU data request with invalid/small length %d\n",
+				data_req->len);
 			break;
 		}
-		msg->l3h = msgb_put(msg, data_req->len);
-		memcpy(msg->l3h, data_req->data, data_req->len);
 
-		if (rsl_imm_assign_cmd(bts, msg->len, msg->data)) {
-			msgb_free(msg);
-			rc = -EIO;
-		}
-		break;
-	case PCU_IF_SAPI_AGCH_DT:
-		/* DT = direct tlli. A tlli is prefixed */
+		pch = (struct gsm_pcu_if_pch *)data_req->data;
+		pag_grp = gsm0502_calc_paging_group(&bts->si_common.chan_desc, str_to_imsi(pch->imsi));
+		gsm48_imm_ass = (struct gsm48_imm_ass *)pch->data;
 
-		if (data_req->len < 5) {
-			LOGP(DPCU, LOGL_ERROR, "Received PCU data request with invalid/small length %d\n",
-			     data_req->len);
-			break;
-		}
-		memcpy(&tlli, data_req->data, 4);
+		if (gsm48_imm_ass->msg_type == GSM48_MT_RR_IMM_ASS)
+			return pcu_rx_rr_imm_ass_pch(bts, pag_grp, pch, pch->confirm);
 
-		msg = msgb_alloc(data_req->len - 4, "pcu_agch");
-		if (!msg) {
-			rc = -ENOMEM;
-			break;
-		}
-		msg->l3h = msgb_put(msg, data_req->len - 4);
-		memcpy(msg->l3h, data_req->data + 4, data_req->len - 4);
-
-		if (bts->type == GSM_BTS_TYPE_RBS2000)
-			rc = rsl_ericsson_imm_assign_cmd(bts, tlli, msg->len, msg->data);
-		else
-			rc = rsl_imm_assign_cmd(bts, msg->len, msg->data);
-
-		if (rc) {
-			msgb_free(msg);
-			rc = -EIO;
-		}
+		if (pcu_rx_rr_paging_pch(bts, pag_grp, pch))
+			return -EIO;
+		if (pch->confirm)
+			return pcu_tx_data_cnf(bts, pch->msg_id, PCU_IF_SAPI_PCH_2);
 		break;
 	default:
-		LOGP(DPCU, LOGL_ERROR, "Received PCU data request with "
+		LOG_BTS(bts, DPCU, LOGL_ERROR, "Received PCU data request with "
 			"unsupported sapi %d\n", data_req->sapi);
 		rc = -EINVAL;
 	}
@@ -480,6 +583,92 @@ static int pcu_rx_data_req(struct gsm_bts *bts, uint8_t msg_type,
 	return rc;
 }
 
+static int pcu_tx_si(const struct gsm_bts *bts, enum osmo_sysinfo_type si_type, bool enable)
+{
+	/* the SI is per-BTS so it doesn't matter which TRX we use */
+	struct gsm_bts_trx *trx = bts->c0;
+
+	uint8_t si_buf[GSM_MACBLOCK_LEN];
+	uint8_t len;
+	int rc;
+
+	if (enable) {
+		memcpy(si_buf, GSM_BTS_SI(bts, si_type), GSM_MACBLOCK_LEN);
+		len = GSM_MACBLOCK_LEN;
+		LOG_BTS(bts, DPCU, LOGL_DEBUG, "Updating SI%s to PCU: %s\n",
+		     get_value_string(osmo_sitype_strs, si_type),
+		     osmo_hexdump_nospc(si_buf, GSM_MACBLOCK_LEN));
+	} else {
+		si_buf[0] = si_type;
+		len = 1;
+
+		/* Note: SI13 is the only system information type that is revked
+		 * by just sending a completely empty message. This is due to
+		 * historical reasons */
+		if (si_type != SYSINFO_TYPE_13)
+			len = 0;
+
+		LOG_BTS(bts, DPCU, LOGL_DEBUG, "Revoking SI%s from PCU\n",
+			get_value_string(osmo_sitype_strs, si_buf[0]));
+	}
+
+	/* The low-level data like FN, ARFCN etc will be ignored but we have to
+	 * set lqual high enough to bypass the check at lower levels */
+	rc = pcu_tx_data_ind(&trx->ts[0], PCU_IF_SAPI_BCCH, 0, 0, 0, si_buf, len,
+			     0, 0, 0, INT16_MAX);
+	if (rc < 0)
+		LOG_BTS(bts, DPCU, LOGL_NOTICE, "Failed to send SI%s to PCU: rc=%d\n",
+			get_value_string(osmo_sitype_strs, si_type), rc);
+
+	return rc;
+}
+
+static int pcu_tx_si_all(struct gsm_bts *bts)
+{
+	const enum osmo_sysinfo_type si_types[] = { SYSINFO_TYPE_1, SYSINFO_TYPE_2, SYSINFO_TYPE_3, SYSINFO_TYPE_13 };
+	unsigned int i;
+	int rc = 0;
+
+	for (i = 0; i < ARRAY_SIZE(si_types); i++) {
+		if (GSM_BTS_HAS_SI(bts, si_types[i])) {
+			rc = pcu_tx_si(bts, si_types[i], true);
+			if (rc < 0)
+				return rc;
+		} else {
+			LOG_BTS(bts, DPCU, LOGL_INFO,
+				"SI%s is not available on PCU connection\n",
+				get_value_string(osmo_sitype_strs, si_types[i]));
+		}
+	}
+
+	return 0;
+}
+
+static int pcu_rx_txt_ind(struct gsm_bts *bts,
+			  const struct gsm_pcu_if_txt_ind *txt)
+{
+	int rc;
+
+	switch (txt->type) {
+	case PCU_VERSION:
+		LOG_BTS(bts, DPCU, LOGL_INFO, "OsmoPCU version %s connected\n",
+		     txt->text);
+		rc = pcu_tx_si_all(bts);
+		if (rc < 0)
+			return -EINVAL;
+		break;
+	case PCU_OML_ALERT:
+		LOG_BTS(bts, DPCU, LOGL_ERROR, "PCU external alarm: %s\n", txt->text);
+		break;
+	default:
+		LOG_BTS(bts, DPCU, LOGL_ERROR, "Unknown TXT_IND type %u received\n",
+		     txt->type);
+		return -EINVAL;
+	}
+
+	return 0;
+}
+
 #define CHECK_IF_MSG_SIZE(prim_len, prim_msg) \
 	do { \
 		size_t _len = PCUIF_HDR_SIZE + sizeof(prim_msg); \
@@ -495,8 +684,9 @@ static int pcu_rx(struct gsm_network *net, uint8_t msg_type,
 	int rc = 0;
 	struct gsm_bts *bts;
 
-	/* FIXME: allow multiple BTS */
-	bts = llist_entry(net->bts_list.next, struct gsm_bts, list);
+	bts = gsm_bts_num(net, pcu_prim->bts_nr);
+	if (!bts)
+		return -EINVAL;
 
 	switch (msg_type) {
 	case PCU_IF_MSG_DATA_REQ:
@@ -504,6 +694,10 @@ static int pcu_rx(struct gsm_network *net, uint8_t msg_type,
 		CHECK_IF_MSG_SIZE(prim_len, pcu_prim->u.data_req);
 		rc = pcu_rx_data_req(bts, msg_type, &pcu_prim->u.data_req);
 		break;
+	case PCU_IF_MSG_TXT_IND:
+		CHECK_IF_MSG_SIZE(prim_len, pcu_prim->u.txt_ind);
+		rc = pcu_rx_txt_ind(bts, &pcu_prim->u.txt_ind);
+		break;
 	default:
 		LOGP(DPCU, LOGL_ERROR, "Received unknown PCU msg type %d\n",
 			msg_type);
@@ -517,9 +711,9 @@ static int pcu_rx(struct gsm_network *net, uint8_t msg_type,
  * PCU socket interface
  */
 
-static int pcu_sock_send(struct gsm_bts *bts, struct msgb *msg)
+static int pcu_sock_send(struct gsm_network *net, struct msgb *msg)
 {
-	struct pcu_sock_state *state = bts->pcu_state;
+	struct pcu_sock_state *state = net->pcu_state;
 	struct osmo_fd *conn_bfd;
 	struct gsm_pcu_if *pcu_prim = (struct gsm_pcu_if *) msg->data;
 
@@ -544,25 +738,10 @@ static int pcu_sock_send(struct gsm_bts *bts, struct msgb *msg)
 	return 0;
 }
 
-static void pcu_sock_close(struct pcu_sock_state *state)
+static void pdch_deact_bts(struct gsm_bts *bts)
 {
-	struct osmo_fd *bfd = &state->conn_bfd;
-	struct gsm_bts *bts;
 	struct gsm_bts_trx *trx;
-	struct gsm_bts_trx_ts *ts;
-	int i, j;
-
-	/* FIXME: allow multiple BTS */
-	bts = llist_entry(state->net->bts_list.next, struct gsm_bts, list);
-
-	LOGP(DPCU, LOGL_NOTICE, "PCU socket has LOST connection\n");
-
-	close(bfd->fd);
-	bfd->fd = -1;
-	osmo_fd_unregister(bfd);
-
-	/* re-enable the generation of ACCEPT for new connections */
-	osmo_fd_read_enable(&state->listen_bfd);
+	int j;
 
 #if 0
 	/* remove si13, ... */
@@ -571,18 +750,38 @@ static void pcu_sock_close(struct pcu_sock_state *state)
 #endif
 
 	/* release PDCH */
-	for (i = 0; i < 8; i++) {
-		trx = gsm_bts_trx_num(bts, i);
-		if (!trx)
-			break;
-		for (j = 0; j < 8; j++) {
-			ts = &trx->ts[j];
+	llist_for_each_entry(trx, &bts->trx_list, list) {
+		for (j = 0; j < ARRAY_SIZE(trx->ts); j++) {
+			struct gsm_bts_trx_ts *ts = &trx->ts[j];
+			/* BSC co-located PCU applies only to Ericsson RBS, which supports only GSM_PCHAN_OSMO_DYN.
+			 * So we need to deact only on this pchan kind. */
 			if (ts->mo.nm_state.operational == NM_OPSTATE_ENABLED
-			    && ts->pchan_is == GSM_PCHAN_PDCH) {
-				printf("l1sap_chan_rel(trx,gsm_lchan2chan_nr(ts->lchan));\n");
+			    && ts->pchan_on_init == GSM_PCHAN_OSMO_DYN) {
+				ts_pdch_deact(ts);
 			}
 		}
 	}
+}
+
+static void pcu_sock_close(struct pcu_sock_state *state)
+{
+	struct osmo_fd *bfd = &state->conn_bfd;
+	struct gsm_bts *bts;
+
+	LOGP(DPCU, LOGL_NOTICE, "PCU socket has LOST connection\n");
+
+	osmo_fd_unregister(bfd);
+	close(bfd->fd);
+	bfd->fd = -1;
+
+	/* re-enable the generation of ACCEPT for new connections */
+	osmo_fd_read_enable(&state->listen_bfd);
+
+	/* Disable all PDCHs on all BTSs that are served by the PCU */
+	llist_for_each_entry(bts, &state->net->bts_list, list) {
+		if (bsc_co_located_pcu(bts))
+			pdch_deact_bts(bts);
+	}
 
 	/* flush the queue */
 	while (!llist_empty(&state->upqueue)) {
@@ -700,36 +899,53 @@ static int pcu_sock_cb(struct osmo_fd *bfd, unsigned int flags)
 	return rc;
 }
 
+static void pdch_act_bts(struct gsm_bts *bts)
+{
+	struct gsm_bts_trx *trx;
+	int j;
+
+	/* activate PDCH */
+	llist_for_each_entry(trx, &bts->trx_list, list) {
+		for (j = 0; j < ARRAY_SIZE(trx->ts); j++) {
+			struct gsm_bts_trx_ts *ts = &trx->ts[j];
+			/* (See comment in pdch_deact_bts above) */
+			if (ts->mo.nm_state.operational == NM_OPSTATE_ENABLED
+			    && ts->pchan_on_init == GSM_PCHAN_OSMO_DYN) {
+				ts_pdch_act(ts);
+			}
+		}
+	}
+}
+
 /* accept connection coming from PCU */
 static int pcu_sock_accept(struct osmo_fd *bfd, unsigned int flags)
 {
 	struct pcu_sock_state *state = (struct pcu_sock_state *)bfd->data;
 	struct osmo_fd *conn_bfd = &state->conn_bfd;
 	struct sockaddr_un un_addr;
+	struct gsm_bts *bts;
 	socklen_t len;
-	int rc;
+	int fd;
 
 	len = sizeof(un_addr);
-	rc = accept(bfd->fd, (struct sockaddr *) &un_addr, &len);
-	if (rc < 0) {
+	fd = accept(bfd->fd, (struct sockaddr *)&un_addr, &len);
+	if (fd < 0) {
 		LOGP(DPCU, LOGL_ERROR, "Failed to accept a new connection\n");
 		return -1;
 	}
 
 	if (conn_bfd->fd >= 0) {
-		LOGP(DPCU, LOGL_NOTICE, "PCU connects but we already have "
-			"another active connection ?!?\n");
+		LOGP(DPCU, LOGL_NOTICE, "PCU connects but we already have another active connection ?!?\n");
 		/* We already have one PCU connected, this is all we support */
 		osmo_fd_read_disable(&state->listen_bfd);
-		close(rc);
+		close(fd);
 		return 0;
 	}
 
-	osmo_fd_setup(conn_bfd, rc, OSMO_FD_READ, pcu_sock_cb, state, 0);
+	osmo_fd_setup(conn_bfd, fd, OSMO_FD_READ, pcu_sock_cb, state, 0);
 
 	if (osmo_fd_register(conn_bfd) != 0) {
-		LOGP(DPCU, LOGL_ERROR, "Failed to register new connection "
-			"fd\n");
+		LOGP(DPCU, LOGL_ERROR, "Failed to register new connection fd\n");
 		close(conn_bfd->fd);
 		conn_bfd->fd = -1;
 		return -1;
@@ -737,11 +953,17 @@ static int pcu_sock_accept(struct osmo_fd *bfd, unsigned int flags)
 
 	LOGP(DPCU, LOGL_NOTICE, "PCU socket connected to external PCU\n");
 
+	/* Activate all PDCHs on all BTSs that are served by the PCU */
+	llist_for_each_entry(bts, &state->net->bts_list, list) {
+		if (bsc_co_located_pcu(bts))
+			pdch_act_bts(bts);
+	}
+
 	return 0;
 }
 
 /* Open connection to PCU */
-int pcu_sock_init(const char *path, struct gsm_bts *bts)
+int pcu_sock_init(struct gsm_network *net)
 {
 	struct pcu_sock_state *state;
 	struct osmo_fd *bfd;
@@ -752,12 +974,12 @@ int pcu_sock_init(const char *path, struct gsm_bts *bts)
 		return -ENOMEM;
 
 	INIT_LLIST_HEAD(&state->upqueue);
-	state->net = bts->network;
+	state->net = net;
 	state->conn_bfd.fd = -1;
 
 	bfd = &state->listen_bfd;
 
-	rc = osmo_sock_unix_init(SOCK_SEQPACKET, 0, path, OSMO_SOCK_F_BIND);
+	rc = osmo_sock_unix_init(SOCK_SEQPACKET, 0, net->pcu_sock_path, OSMO_SOCK_F_BIND);
 	if (rc < 0) {
 		LOGP(DPCU, LOGL_ERROR, "Could not create unix socket: %s\n",
 			strerror(errno));
@@ -776,16 +998,17 @@ int pcu_sock_init(const char *path, struct gsm_bts *bts)
 		return rc;
 	}
 
-	LOGP(DPCU, LOGL_INFO, "Started listening on PCU socket: %s\n", path);
+	LOGP(DPCU, LOGL_INFO, "Started listening on PCU socket (PCU IF v%u): %s\n",
+	     PCU_IF_VERSION, net->pcu_sock_path);
 
-	bts->pcu_state = state;
+	net->pcu_state = state;
 	return 0;
 }
 
 /* Close connection to PCU */
-void pcu_sock_exit(struct gsm_bts *bts)
+void pcu_sock_exit(struct gsm_network *net)
 {
-	struct pcu_sock_state *state = bts->pcu_state;
+	struct pcu_sock_state *state = net->pcu_state;
 	struct osmo_fd *bfd, *conn_bfd;
 
 	if (!state)
@@ -795,8 +1018,8 @@ void pcu_sock_exit(struct gsm_bts *bts)
 	if (conn_bfd->fd > 0)
 		pcu_sock_close(state);
 	bfd = &state->listen_bfd;
-	close(bfd->fd);
 	osmo_fd_unregister(bfd);
+	close(bfd->fd);
 	talloc_free(state);
-	bts->pcu_state = NULL;
+	net->pcu_state = NULL;
 }
